/*
 * Copyright (c) 2020 Andri Yngvason
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include <aml.h>
#include <signal.h>

static void on_line(void* handler)
{
	char line[256];
	fscanf(stdin, "%s", line);
	
	printf("Got line: %s\n", line);

	if (strncmp(line, "exit", sizeof(line)) == 0)
		aml_exit(aml_get_default());
}

static void on_sigint(void* sig)
{
	aml_exit(aml_get_default());
}

int main()
{
	struct aml* aml = aml_new();
	if (!aml)
		return 1;

	aml_set_default(aml);

	struct aml_signal* sig = aml_signal_new(SIGINT, on_sigint, NULL, NULL);
	if (!sig)
		goto failure;

	aml_start(aml, sig);
	aml_unref(sig);

	struct aml_handler* handler =
		aml_handler_new(fileno(stdin), on_line, NULL, NULL);
	if (!handler)
		goto failure;

	aml_start(aml, handler);
	aml_unref(handler);

	aml_run(aml);

	printf("Exiting...\n");

	aml_unref(aml);
	return 0;

failure:
	aml_unref(aml);
	return 1;
}
